function [omega_A omega_B t k like div] = dnds_ov(piA,piB,hexa_piA,hexa_piB,nijA,nijB)
% Version 1.0 11/2008 by Niv Sabath email:nsabath@uh.edu
% Estimating selection intensities in overlapping genes
%
% input:
% seq1, seq2: two gap-less aligned homologous DNA sequences of overlapping genes.
% the sequences must be in the reading-frame of the reference gene (gene A)
% orientation: 1 - same-strand; 2 - opposite-strand
% phase: 0 (only in opposite-strand overlaps), 1, or 2
% non_ov_a1, non_ov_a2 (optional): two aligned homologous DNA sequences of the non-overlapping
% regions of the reference gene (gene A). [] if not exist.  
% non_b1, non_ov_b2 (optional): two aligned homologous DNA sequences of the non-overlapping
% regions of gene B. [] if not exist.  
% ALL sequences must be of complete codons - mod(length,3) == 0 
%
% output:
% omega_A: estimated selection intensity on gene A 
% omega_B: estimated selection intensity on gene B
% t: estimated time substitutions/codon 
% k: estimated transition/transversion rate ratio 
% like: the likelihood 
% div: % divergence of seq1 and seq2 
%
% Requires Matlab Optimization Toolbox
%
% example:
% load 'example.mat'
% [omega_A omega_B t k like div] = 
%    dnds_ov(seq1,seq2,orientation,phase,non_ov_a1,non_ov_a2,non_ov_b1,non_ov_b2);


% check input parameters:

%if(orientation ~= 1 && orientation ~= 2 )
%     error('dnds_ov:Invalid Orientation',...
%      'Enter 1 for same-strand; 2 for opposite-strand');
%end

%if(phase < 0 || phase > 2 )
%     error('dnds_ov:Invalid Phase',...
%      'Enter 1 or 2 for same-strand; 0, 1, or 2 for opposite-strand');
%end

%if(orientation == 1 && phase == 0)
%     error('dnds_ov:Invalid Orientation and/or phase',...
%      'Same-strand overlaps in phase 0 have the same selection constraints and are not accepted by the program');
%end

%if(length(seq1) < 30 || length(seq2) < 30)
%     error('dnds_ov:Sequence too short',...
%      'dnds_ov needs at least 10 codons in each sequence');
%end

%if(length(seq1) ~= length(seq2))
%     error('dnds_ov:Sequences have different length',...
%      'Enter gap-less aligned homologous DNA sequences');
%end

%s1 = mynt2int(seq1);
%s2 = mynt2int(seq2);

%div = length(find(s1-s2))/length(s1);

%if div < 0.05
%    'Warning: low divergence:'
%    div
%    'Estimation may be inaccurate'
%end

%if(isempty(non_ov_a1) || isempty(non_ov_a2))
%    'Warning: No non-overlap region for gene A, use the overlap region to calculate codon frequencies'
%    a1 = s1;
%    a2 = s2;
%else
%    a1 = mynt2int(non_ov_a1);
%    a2 = mynt2int(non_ov_a2);
%end

%if(isempty(non_ov_b1) || isempty(non_ov_b2))
%    'Warning: No non-overlap region for gene B, use the overlap region to calculate codon frequencies'
%    if orientation == 1 % same-strand
%        b1 = s1(1+phase:end-3+phase);
%        b2 = s2(1+phase:end-3+phase);
%    else % opposite-strand
%        b1 = revcom(s1(1+phase:end-3+phase));
%        b2 = revcom(s2(1+phase:end-3+phase));
%    end
%else
%    b1 = mynt2int(non_ov_b1);
%    b2 = mynt2int(non_ov_b2);
%end

%ll = [length(s1) length(a1) length(a2) length(b1) length(b2)]; 
%ll_mod = mod(ll,3);
%if any(ll_mod)
%   error('dnds_ov:Bad Sequences',...
%      'ALL sequences must be of complete codons - mod(length,3) == 0');
%end  

% check if the sequences contain stop codons
%if  orientation == 1 % same-strand
%    st_a = [find_stops(s1), find_stops(s2)];
%    st_b = [find_stops(s1(1+phase:end-3+phase)), find_stops(s2(1+phase:end-3+phase))];
%else
%    st_a = [find_stops(s1), find_stops(s2)];
%    st_b = [find_rcstops(s1(1+phase:end-3+phase)), find_rcstops(s2(1+phase:end-3+phase))];
%end

%if any(st_a)
%    'Warning: stop codons in gene A'
%end
%if any(st_b)
%    'Warning: stop codons in gene B'
%end

% run the program
% 
% if orientation == 1 % same-strand
%     if phase==1
[omega_A,omega_B,t,k,like] = dndsml_phase1_ss(piA,piB,hexa_piA,hexa_piB,nijA,nijB);
%     elseif phase==2
%         [omega_A omega_B t k like] = dndsml_phase2_ss(s1,s2,a1,a2,b1,b2);
%     end
% else % opposite-strand
%     if phase==0
%         [omega_A omega_B t k like] = dndsml_phase0_opp(s1,s2,a1,a2,b1,b2);
%     elseif phase==1
%         [omega_A omega_B t k like] = dndsml_phase1_opp(s1,s2,a1,a2,b1,b2);
%     elseif phase==2
%         [omega_A omega_B t k like] = dndsml_phase2_opp(s1,s2,a1,a2,b1,b2);
%     end
% end

vv = [omega_A omega_B t k];
if length(find(vv==0.001)) > 0 || length(find(vv==50)) > 0 || t == 5
    'Warning: Likelihood search did not converge. Estimation may be inaccurate'
end
fprintf('%f\t%f\t%f\t%f\n',omega_A,omega_B,t,k)

